/*
 * sdramif.v -- simple interface for SDRAM to 6502 bus.
 * 
 * assumes 8 bit 6502 bus, and 16 bit SDRAM bus. Only uses bits 7-0 of the SDRAM.
 *
 * (C) Arlet Ottens, <arlet@c-scape.nl>
 */

module sdramif(
	input clk,
 	input [15:0] AB,
	input WE,
	input [7:0] DO,
	output wor [7:0] DI,
	output RDY,
	output reg [7:0] DI_CTRL,
	input ctrl,	// window control

	input [23:0] vid_addr,
	output [15:0] vid_rd_data,
	input vid_rd_req,
	output vid_rd_ack,
	output vid_rd_valid,

	output reg [23:0] sdram_addr,
	input [15:0] sdram_rd_data,
	input sdram_next_rd_valid,
	input sdram_rd_valid,
	input sdram_rd_ack,
	output reg sdram_rd_req,

	output reg [15:0] sdram_wr_data,
	input sdram_next_wr_ack,
	output reg sdram_wr_req
	
	);

reg [15:0] ABH = 0;

wire sel_yuv = (AB[15:8] == 8'ha0);	

parameter 
	IDLE    = 3'd0,
	READ1	= 3'd1,
	READ2 	= 3'd2,
	WRITE1  = 3'd3,
	VIDEO   = 3'd4;

reg [2:0] state = IDLE;
reg [2:0] next;
reg di_enable = 0;

// high byte for SDRAM access 
reg [7:0] DOH;

always @(posedge clk) 
    if( ctrl & WE )
	case( AB[1:0] )
	    0: ABH[ 7:0] <= DO;
	    1: ABH[15:8] <= DO;
	    2: DOH	 <= DO;
	endcase

always @*
    case( AB[1:0] )
        0: DI_CTRL <= ABH[ 7:0];
        1: DI_CTRL <= ABH[15:8];
	2: DI_CTRL <= DOH; 
  default: DI_CTRL <= 0;
    endcase

always @(posedge clk) 
    di_enable <= sel_yuv;

always @(posedge clk) 
    state <= next;

always @* begin
    next = state;
    case( state )
        IDLE: 	if( sel_yuv )
			if( WE )		next = WRITE1;
			else			next = READ1;
		else if( vid_rd_req )		next = VIDEO;

	READ1:	if( sdram_next_rd_valid )	next = IDLE;
		else if( sdram_rd_ack )		next = READ2;

	READ2:  if( sdram_next_rd_valid )	next = IDLE;

        WRITE1: if( sdram_next_wr_ack )		next = IDLE;

	VIDEO : if( ~vid_rd_req )
		    if( sel_yuv )
			if( WE )		next = WRITE1;
			else			next = READ1;
		    else			next = IDLE;
    endcase
end

reg busy = 0;

always @(posedge clk)
    if( (next == VIDEO) && vid_rd_req )
        busy <= 1;
    else if( sdram_rd_ack )
        busy <= 0;

always @(posedge clk) begin
    if( next == VIDEO && (~busy | sdram_rd_ack) )
	sdram_addr <= vid_addr;
    else if( next == WRITE1 ) begin
	sdram_wr_data <= { DOH, DO };
	sdram_addr <= { ABH, AB[7:0] };
    end else if( next == READ1 ) begin
        sdram_addr <= { ABH, AB[7:0] };
    end
    sdram_rd_req <= (next == READ1 || next == VIDEO);
    sdram_wr_req <= (next == WRITE1);
end
	 
assign RDY = ~sel_yuv || (next == IDLE );
assign DI = di_enable ?  sdram_rd_data[7:0] : 0;

wire channel;

`ifdef SIM
fifo1 fifo1(
`else
fifo #(1) fifo(
`endif
	.clk(clk),
	.rd(sdram_rd_valid),
	.wr(sdram_rd_ack),
	.in(state == VIDEO),
	.out(channel) );

assign vid_rd_ack = ((next == VIDEO) & ~busy) || ((state == VIDEO) & sdram_rd_ack & vid_rd_req);
assign vid_rd_valid = channel & sdram_rd_valid;
assign vid_rd_data = sdram_rd_data;

endmodule
